"""Tests for the GraphQL interface for gbp-ps"""

# pylint: disable=missing-docstring

from typing import Any

from django.test.client import Client
from unittest_fixtures import requires

from . import TestCase, make_build_process


def graphql(query: str, variables: dict[str, Any] | None = None) -> Any:
    """Execute GraphQL query on the Django test client.

    Return the parsed JSON response
    """
    client = Client()
    response = client.post(
        "/graphql",
        {"query": query, "variables": variables},
        content_type="application/json",
    )

    return response.json()


class GetProcessesTests(TestCase):
    query = """
    {
      buildProcesses {
        machine
        id
        buildHost
        package
        phase
         startTime
      }
    }
    """

    def test_empty(self) -> None:
        result = graphql(self.query)

        self.assertNotIn("errors", result)
        self.assertEqual(result["data"]["buildProcesses"], [])

    def test_nonempty(self) -> None:
        build_process = make_build_process()

        result = graphql(self.query)

        self.assertNotIn("errors", result)
        self.assertEqual(result["data"]["buildProcesses"], [build_process.to_dict()])

    def test_non_empty_with_final_processes(self) -> None:
        live_process = make_build_process(package="sys-apps/systemd-254.5-r1")
        make_build_process(package="sys-libs/efivar-38", phase="clean")

        result = graphql(self.query)

        self.assertNotIn("errors", result)
        self.assertEqual(result["data"]["buildProcesses"], [live_process.to_dict()])

    def test_non_empty_with_final_processes_included(self) -> None:
        make_build_process(package="sys-apps/systemd-254.5-r1")
        make_build_process(package="sys-libs/efivar-38", phase="clean")
        query = "{buildProcesses(includeFinal: true) { machine id package startTime }}"

        result = graphql(query)

        self.assertNotIn("errors", result)
        self.assertEqual(len(result["data"]["buildProcesses"]), 2)


@requires("repo")
class AddBuildProcessesTests(TestCase):
    query = """
    mutation (
      $process: BuildProcessInput!,
    ) {
      addBuildProcess(
        process: $process,
      ) {
        message
      }
    }
    """

    def test(self) -> None:
        process = make_build_process()
        result = graphql(self.query, {"process": process.to_dict()})

        self.assertNotIn("errors", result)
        processes = [*self.fixtures.repo.get_processes()]
        self.assertEqual(processes, [process])

    def test_update(self) -> None:
        p_dict = make_build_process().to_dict()
        graphql(self.query, {"process": p_dict})

        p_dict["phase"] = "postinst"
        result = graphql(self.query, {"process": p_dict})
        self.assertNotIn("errors", result)
        processes = [*self.fixtures.repo.get_processes()]
        self.assertEqual(processes[0].phase, "postinst")

    def test_empty_phase_does_not_get_added(self) -> None:
        p_dict = make_build_process(phase="", add_to_repo=False).to_dict()
        result = graphql(self.query, {"process": p_dict})

        self.assertNotIn("errors", result)
        self.assertEqual([*self.fixtures.repo.get_processes(include_final=True)], [])

    def test_empty_machine_does_not_get_added(self) -> None:
        p_dict = make_build_process(machine="", add_to_repo=False).to_dict()
        result = graphql(self.query, {"process": p_dict})

        self.assertNotIn("errors", result)
        self.assertEqual([*self.fixtures.repo.get_processes(include_final=True)], [])



"""Django RepositoryType"""

from typing import Iterable

from gbp_ps.exceptions import RecordAlreadyExists, RecordNotFoundError
from gbp_ps.settings import Settings
from gbp_ps.types import BuildProcess


class DjangoRepository:
    """Django ORM-based BuildProcess repository"""

    def __init__(self, _settings: Settings) -> None:
        # pylint: disable=import-outside-toplevel
        from gbp_ps.models import BuildProcess as BuildProcessModel

        self.model: type[BuildProcessModel] = BuildProcessModel

    def add_process(self, process: BuildProcess) -> None:
        """Add the given BuildProcess to the repository

        If the process already exists in the repo, RecordAlreadyExists is raised
        """
        # pylint: disable=import-outside-toplevel
        import django.db.utils
        from django.db.models import Q

        # If this package exists in another build, remove it. This (usually) means the
        # other build failed
        self.model.objects.filter(
            ~Q(build_id=process.build_id),
            machine=process.machine,
            package=process.package,
        ).delete()

        build_process_model = self.model.from_object(process)

        try:
            build_process_model.save()
        except django.db.utils.IntegrityError:
            raise RecordAlreadyExists(process) from None

    def update_process(self, process: BuildProcess) -> None:
        """Update the given build process

        Only updates the phase field

        If the build process doesn't exist in the repo, RecordNotFoundError is raised.
        """
        try:
            build_process_model = self.model.objects.get(
                machine=process.machine,
                build_id=process.build_id,
                package=process.package,
            )
        except self.model.DoesNotExist:
            raise RecordNotFoundError(process) from None

        build_process_model.to_object().ensure_updateable(process)

        build_process_model.phase = process.phase
        build_process_model.build_host = process.build_host
        build_process_model.save()

    def get_processes(self, include_final: bool = False) -> Iterable[BuildProcess]:
        """Return the process records from the repository

        If include_final is True also include processes in their "final" phase. The
        default value is False.
        """
        query = self.model.objects.order_by("start_time")
        if not include_final:
            query = query.exclude(phase__in=BuildProcess.final_phases)

        return (model.to_object() for model in query)



"""Show dump an appropriate /etc/portage/bashrc for gbp-ps"""

import argparse
import subprocess as sp
from importlib import resources
from typing import Callable

from gbpcli.gbp import GBP
from gbpcli.types import Console

BASHRC_FILENAME = "bashrc.bash"
LOCAL_BASHRC_FILENAME = "bashrc-local.bash"


def handler(args: argparse.Namespace, gbp: GBP, console: Console) -> int:
    """Show dump an appropriate /etc/portage/bashrc for gbp-ps"""
    formatter: Callable[[str, GBP], str]
    if args.local:
        filename = LOCAL_BASHRC_FILENAME
        formatter = format_local
    else:
        filename = BASHRC_FILENAME
        formatter = format_bashrc

    with resources.open_text("gbp_ps.assets", filename, encoding="utf-8") as fp:
        bashrc = fp.read()

    bashrc = formatter(bashrc, gbp)

    console.out.print(bashrc)

    return 0


def parse_args(parser: argparse.ArgumentParser) -> None:
    """Parse command-line arguments"""
    parser.add_argument(
        "--local",
        "-l",
        action="store_true",
        default=False,
        help="Generate a bashrc for local ebuild processes",
    )


def format_bashrc(bashrc: str, gbp: GBP) -> str:
    """Format the given GPB node's bashrc"""
    # pylint: disable=protected-access
    return bashrc.replace("http://gbp/graphql", gbp.query._url)


def format_local(bashrc: str, _gbp: GBP) -> str:
    """Format the given local bashrc"""
    return bashrc.replace("/var/tmp", portage_tmpdir())


def portage_tmpdir() -> str:
    """Get the local PORTAGE_TMPDIR

    If this cannot be determined, return the default ("/var/tmp")
    """
    with sp.Popen(["portageq", "envvar", "PORTAGE_TMPDIR"], stdout=sp.PIPE) as proc:
        stdout = proc.stdout
        assert stdout
        output = stdout.read().decode("utf-8")

    if proc.wait() == 0:
        return output.strip()
    return "/var/tmp"



"""Redis RepositoryType"""

import datetime as dt
import json
from typing import Iterable

import redis

from gbp_ps.exceptions import RecordAlreadyExists, RecordNotFoundError
from gbp_ps.settings import Settings
from gbp_ps.types import BuildProcess

ENCODING = "UTF-8"


class RedisRepository:
    """Redis backend for the process table"""

    def __init__(self, settings: Settings) -> None:
        self._redis = redis.Redis.from_url(settings.REDIS_URL)
        self._key = settings.REDIS_KEY
        self.time = settings.REDIS_KEY_EXPIRATION

    def key(self, process: BuildProcess) -> bytes:
        """Return the redis key for the given BuildProcess"""
        return f"{self._key}:{process.machine}:{process.package}:{process.build_id}".encode(
            ENCODING
        )

    def value(self, process: BuildProcess) -> bytes:
        """Return the redis value for the given BuildProcess"""
        return json.dumps(
            {
                "build_host": process.build_host,
                "phase": process.phase,
                "start_time": process.start_time.isoformat(),
            }
        ).encode(ENCODING)

    def process_to_redis(self, process: BuildProcess) -> tuple[bytes, bytes]:
        """Return the redis key and value for the given BuildProcess"""
        return self.key(process), self.value(process)

    def redis_to_process(self, key: bytes, value: bytes) -> BuildProcess:
        """Convert the given key and value to a BuildProcess"""
        machine, package, build_id = key.decode(ENCODING).split(":")[1:]
        data = json.loads(value.decode(ENCODING))

        return BuildProcess(
            build_host=data["build_host"],
            build_id=build_id,
            machine=machine,
            package=package,
            phase=data["phase"],
            start_time=dt.datetime.fromisoformat(data["start_time"]),
        )

    def add_process(self, process: BuildProcess) -> None:
        """Add the given BuildProcess to the repository

        If the process already exists in the repo, RecordAlreadyExists is raised
        """
        # If this package exists in another build, remove it. This (usually) means the
        # other build failed
        self.delete_existing_processes(process)
        key, value = self.process_to_redis(process)
        previous = self._redis.get(key)

        if previous and self.redis_to_process(key, previous).is_same_as(process):
            raise RecordAlreadyExists(process)

        self._redis.setex(key, self.time, value)

    def delete_existing_processes(self, process: BuildProcess) -> int:
        """Delete existing processes like process

        By "existing" we mean processes in Redis that have the same machine and package
        but different build_id.

        Return the number of processes deleted.
        """
        build_id = process.build_id.encode(ENCODING)
        deleted_count = 0
        pattern = f"{self._key}:{process.machine}:{process.package}:*".encode(ENCODING)
        for key in self._redis.keys(pattern):
            if key.split(b":")[3] != build_id:
                self._redis.delete(key)
                deleted_count += 1
        return deleted_count

    def update_process(self, process: BuildProcess) -> None:
        """Update the given build process

        Only updates the phase field

        If the build process doesn't exist in the repo, RecordNotFoundError is raised.
        """
        key = self.key(process)
        previous_value = self._redis.get(key)

        if previous_value is None:
            raise RecordNotFoundError(process)

        self.redis_to_process(key, previous_value).ensure_updateable(process)
        new_value = {
            **json.loads(previous_value),
            **{"phase": process.phase, "build_host": process.build_host},
        }
        self._redis.setex(key, self.time, json.dumps(new_value).encode(ENCODING))

    def get_processes(self, include_final: bool = False) -> Iterable[BuildProcess]:
        """Return the process records from the repository

        If include_final is True also include processes in their "final" phase. The
        default value is False.
        """
        processes = []

        for key in self._redis.keys(f"{self._key}:*".encode(ENCODING)):
            if value := self._redis.get(key):
                process = self.redis_to_process(key, value)

                if include_final or not process.is_finished():
                    processes.append(process)

        processes.sort(key=lambda process: process.start_time)
        return processes



# pylint: disable=missing-docstring
import io
import os
import tempfile
from typing import Mapping
from unittest import mock

import rich.console
from django.test.client import Client
from gbpcli.gbp import GBP
from gbpcli.theme import get_theme_from_string
from gbpcli.types import Console
from requests import PreparedRequest, Response
from requests.adapters import BaseAdapter
from requests.structures import CaseInsensitiveDict
from unittest_fixtures import FixtureContext, FixtureOptions, Fixtures, depends

from gbp_ps.repository import Repo, RepositoryType, sqlite
from gbp_ps.settings import Settings

COUNTER = 0


def tempdir(_options: FixtureOptions, _fixtures: Fixtures) -> FixtureContext[str]:
    with tempfile.TemporaryDirectory() as tempdir_:
        yield tempdir_


@depends("tempdir")
def environ(
    options: FixtureOptions, fixtures: Fixtures
) -> FixtureContext[dict[str, str]]:
    new_environ = options.get("environ", {}).copy()
    new_environ["GBP_PS_SQLITE_DATABASE"] = f"{fixtures.tempdir}/db.sqlite"
    with mock.patch.dict("os.environ", new_environ, clear=True):
        yield new_environ


@depends("environ")
def settings(_options: FixtureOptions, _fixtures: Fixtures) -> Settings:
    return Settings.from_environ()


@depends("settings")
def repo(_options: FixtureOptions, fixtures: Fixtures) -> RepositoryType:
    return Repo(fixtures.settings)


def gbp(options: FixtureOptions, _fixtures: Fixtures) -> GBP:
    url = options.get("gbp", {}).get("url", "http://gbp.invalid/")
    gbp_ = GBP(url)
    gbp_.query._session.mount(  # pylint: disable=protected-access
        url, DjangoToRequestsAdapter()
    )

    return gbp_


@depends("tempdir")
def tempdb(_options: FixtureOptions, fixtures: Fixtures) -> str:
    return f"{fixtures.tempdir}/processes.db"


@depends(tempdb)
def repo_fixture(
    _options: FixtureOptions, fixtures: Fixtures
) -> sqlite.SqliteRepository:
    return sqlite.SqliteRepository(Settings(SQLITE_DATABASE=fixtures.tempdb))


class DjangoToRequestsAdapter(BaseAdapter):  # pylint: disable=abstract-method
    """Requests Adapter to call Django views"""

    def send(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        request: PreparedRequest,
        stream: bool = False,
        timeout: None | float | tuple[float, float] | tuple[float, None] = None,
        verify: bool | str = True,
        cert: None | bytes | str | tuple[bytes | str, bytes | str] = None,
        proxies: Mapping[str, str] | None = None,
    ) -> Response:
        assert request.method is not None
        django_response = Client().generic(
            request.method,
            request.path_url,
            data=request.body,
            content_type=request.headers["Content-Type"],
            **request.headers,
        )

        requests_response = Response()
        requests_response.raw = io.BytesIO(django_response.content)
        requests_response.raw.seek(0)
        requests_response.status_code = django_response.status_code
        requests_response.headers = CaseInsensitiveDict(django_response.headers)
        requests_response.encoding = django_response.get("Content-Type", None)
        requests_response.url = str(request.url)
        requests_response.request = request

        return requests_response


@depends()
def console(_options: FixtureOptions, _fixtures: Fixtures) -> FixtureContext[Console]:
    outfile = io.StringIO()
    errfile = io.StringIO()
    theme = get_theme_from_string(os.getenv("GBPCLI_COLORS", ""))
    out = rich.console.Console(
        file=outfile, width=88, theme=theme, highlight=False, record=True
    )
    err = rich.console.Console(file=errfile, width=88, record=True)
    c = Console(out=out, err=err)

    yield c

    # pylint: disable=no-member,global-statement
    if "SAVE_VIRTUAL_CONSOLE" in os.environ and c.out.file.getvalue():
        global COUNTER

        COUNTER += 1
        filename = f"{COUNTER}.svg"
        c.out.save_svg(filename, title="Gentoo Build Publisher")



"""Django models for gbp_ps"""

from __future__ import annotations

from typing import TypeVar

from django.db import models

from gbp_ps.types import BuildProcess as BuildProcessObject

T = TypeVar("T", bound="BuildProcess")


class BuildProcess(models.Model):
    """A BuildProcess record in the database"""

    machine = models.CharField(max_length=255, db_index=True)
    build_id = models.CharField(max_length=255)
    build_host = models.CharField(max_length=255)
    package = models.CharField(max_length=255)
    phase = models.CharField(max_length=255, db_index=True)
    start_time = models.DateTimeField()

    class Meta:
        unique_together = [["machine", "build_id", "build_host", "package"]]

    def to_object(self) -> BuildProcessObject:
        """Convert to the non-ORM object"""
        return BuildProcessObject(
            machine=self.machine,
            build_id=self.build_id,
            build_host=self.build_host,
            package=self.package,
            phase=self.phase,
            start_time=self.start_time,
        )

    @classmethod
    def from_object(cls: type[T], obj: BuildProcessObject) -> T:
        """Convert the non-ORM object to an ORM model"""
        return cls(
            machine=obj.machine,
            build_id=obj.build_id,
            build_host=obj.build_host,
            package=obj.package,
            phase=obj.phase,
            start_time=obj.start_time,
        )



"""gbp-ps tests"""

# pylint: disable=missing-docstring
import argparse
import datetime as dt
import shlex
from functools import wraps
from typing import Any, Callable, Iterable, TypeVar

import gbpcli
from django.test import TestCase as DjangoTestCase
from gbpcli.config import Config
from unittest_fixtures import BaseTestCase

from gbp_ps.repository import Repo, add_or_update_process
from gbp_ps.settings import Settings
from gbp_ps.types import BuildProcess

LOCAL_TIMEZONE = dt.timezone(dt.timedelta(days=-1, seconds=61200), "PDT")


class TestCase(DjangoTestCase, BaseTestCase):
    """Custom TestCase for gbp-ps tests"""


def make_build_process(**kwargs: Any) -> BuildProcess:
    """Create (and save) a BuildProcess"""
    settings = Settings.from_environ()
    add_to_repo = kwargs.pop("add_to_repo", True)
    update_repo = kwargs.pop("update_repo", False)
    attrs: dict[str, Any] = {
        "build_host": "jenkins",
        "build_id": "1031",
        "machine": "babette",
        "package": "sys-apps/systemd-254.5-r1",
        "phase": "compile",
        "start_time": dt.datetime(2023, 11, 11, 12, 20, 52, tzinfo=dt.timezone.utc),
    }
    attrs.update(**kwargs)
    build_process = BuildProcess(**attrs)

    if add_to_repo:
        repo = Repo(settings)
        if update_repo:
            add_or_update_process(repo, build_process)
        else:
            repo.add_process(build_process)

    return build_process


def parse_args(cmdline: str) -> argparse.Namespace:
    args = shlex.split(cmdline)
    parser = gbpcli.build_parser(Config(url="http://gbp.invalid/"))

    return parser.parse_args(args[1:])


T = TypeVar("T")


def parametrized(lists_of_args: Iterable[Iterable[Any]]) -> Callable[..., Any]:
    """Parameterized test"""

    def dec(func: Callable[..., Any]) -> Callable[[TestCase, T], None]:
        @wraps(func)
        def wrapper(self: TestCase, *args: Any, **kwargs: Any) -> None:
            for list_of_args in lists_of_args:
                name = ",".join(str(i) for i in list_of_args)
                with self.subTest(name):
                    func(self, *args, *list_of_args, **kwargs)

        return wrapper

    return dec



# pylint: disable=missing-docstring,too-few-public-methods
import datetime as dt

import factory

from gbp_ps.types import BuildProcess

PACKAGES = (
    "media-libs/tiff-4.7.0",
    "app-misc/pax-utils-1.3.8",
    "media-libs/x265-3.6-r1",
    "sys-fs/cryptsetup-2.7.5-r1",
    "sys-devel/gcc-14.2.1_p20240921",
    "sys-fs/cryptsetup-2.7.5",
)


class BuildProcessFactory(factory.Factory):
    class Meta:
        model = BuildProcess

    machine = "babette"
    build_id = factory.Sequence(str)
    build_host = "builder"
    package = factory.Iterator(PACKAGES)
    phase = factory.Iterator(BuildProcess.build_phases)
    start_time = factory.LazyFunction(
        lambda: dt.datetime.now(tz=dt.UTC).replace(microsecond=0)
    )



"""Database Repository for build processes"""

from __future__ import annotations

import importlib.metadata
from collections.abc import Iterable
from typing import Generator, Protocol

from gbp_ps.exceptions import RecordNotFoundError, UpdateNotAllowedError
from gbp_ps.settings import Settings
from gbp_ps.types import BuildProcess

BACKENDS = {ep.name: ep for ep in importlib.metadata.entry_points(group="gbp_ps.repos")}


class RepositoryType(Protocol):
    """BuildProcess Repository"""

    def __init__(self, settings: Settings) -> None:
        """Initializer"""

    def add_process(self, process: BuildProcess) -> None:
        """Add the given BuildProcess to the repository

        If the process already exists in the repo, RecordAlreadyExists is raised
        """

    def update_process(self, process: BuildProcess) -> None:
        """Update the given build process

        Only updates the phase field

        If the build process doesn't exist in the repo, RecordNotFoundError is raised.
        """

    def get_processes(self, include_final: bool = False) -> Iterable[BuildProcess]:
        """Return the process records from the repository

        If include_final is True also include processes in their "final" phase. The
        default value is False.
        """


def Repo(settings: Settings) -> RepositoryType:  # pylint: disable=invalid-name
    """Return a Repository

    If the GBP_PS_REDIS_URL environment variable is defined and non-empty, return the
    RedisRepository. Otherwise the DjangoRepository is returned.
    """
    if entry_point := BACKENDS.get(settings.STORAGE_BACKEND):
        cls: type[RepositoryType] = entry_point.load()
        return cls(settings)

    raise ValueError(f"Invalid storage backend: {settings.STORAGE_BACKEND!r}")


def add_or_update_process(repo: RepositoryType, process: BuildProcess) -> None:
    """Add or update the process

    Adds the process to the process table. If the process already exists, does an
    update.

    If the update is not allowed (e.g. the previous build host is attempting to finalize
    the process) update is not ignored.
    """
    try:
        repo.update_process(process)
    except RecordNotFoundError:
        repo.add_process(process)
    except UpdateNotAllowedError:
        pass



"""Tests for gbp-ps signal handlers"""

# pylint: disable=missing-docstring
import datetime as dt
from unittest import mock

from gentoo_build_publisher.signals import dispatcher
from gentoo_build_publisher.types import Build
from unittest_fixtures import requires

from gbp_ps import signals
from gbp_ps.types import BuildProcess

from . import TestCase, factories

NODE = "wopr"
START_TIME = dt.datetime(2023, 12, 10, 13, 53, 46, tzinfo=dt.UTC)
BUILD = Build(machine="babette", build_id="10")


@requires("repo")
@mock.patch("gbp_ps.signals._NODE", new=NODE)
@mock.patch("gbp_ps.signals._now", mock.Mock(return_value=START_TIME))
class SignalsTest(TestCase):
    def test_create_build_process(self) -> None:
        process = signals.build_process(BUILD, NODE, "test", START_TIME)

        expected: BuildProcess = factories.BuildProcessFactory(
            build_id=BUILD.build_id,
            build_host=NODE,
            machine=BUILD.machine,
            package="pipeline",
            phase="test",
            start_time=START_TIME,
        )
        self.assertEqual(process, expected)

    def test_prepull_handler(self) -> None:
        signals.prepull_handler(build=BUILD)

        processes = [*self.fixtures.repo.get_processes(include_final=True)]
        expected = signals.build_process(BUILD, NODE, "pull", START_TIME)
        self.assertEqual(processes, [expected])

    def test_postpull_handler(self) -> None:
        signals.postpull_handler(build=BUILD)

        processes = [*self.fixtures.repo.get_processes(include_final=True)]
        expected = signals.build_process(BUILD, NODE, "clean", START_TIME)
        self.assertEqual(processes, [expected])

    def test_handler_updates(self) -> None:
        signals.prepull_handler(build=BUILD)
        signals.postpull_handler(build=BUILD)

        processes = [*self.fixtures.repo.get_processes(include_final=True)]
        expected = signals.build_process(BUILD, NODE, "clean", START_TIME)
        self.assertEqual(processes, [expected])

    def test_dispatcher_calls_prepull_handler(self) -> None:
        dispatcher.emit("prepull", build=BUILD)

        processes = [*self.fixtures.repo.get_processes(include_final=True)]
        expected = signals.build_process(BUILD, NODE, "pull", START_TIME)
        self.assertEqual(processes, [expected])

    def test_dispatcher_calls_postpull_handler(self) -> None:
        dispatcher.emit("postpull", build=BUILD)

        processes = [*self.fixtures.repo.get_processes(include_final=True)]
        expected = signals.build_process(BUILD, NODE, "clean", START_TIME)
        self.assertEqual(processes, [expected])



"""gbp-ps exceptions"""

from functools import wraps
from typing import Any, Callable, ParamSpec, TypeVar

T = TypeVar("T")
P = ParamSpec("P")
RETURN_EXCEPTION = object()


class GBPPSException(Exception):
    """Base exception for gbp-ps"""


class RecordNotFoundError(GBPPSException, LookupError):
    """Raised when a record is not found in the repository"""


class RecordAlreadyExists(GBPPSException):
    """Raised when adding a record that already exists in the repository"""


class UpdateNotAllowedError(GBPPSException):
    """Raised when an update is not allowed"""


def swallow_exception(
    *exceptions: type[BaseException], returns: Any = RETURN_EXCEPTION
) -> Callable[[Callable[P, T]], Callable[P, T | Any]]:
    """Swallow the given exceptions"""

    def decorator(func: Callable[P, T]) -> Callable[P, T | Any]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T | Any:
            try:
                return func(*args, **kwargs)
            except exceptions as exception:
                if returns is RETURN_EXCEPTION:
                    return exception
                return returns

        return wrapper

    return decorator



"""CLI unit tests for gbp-ps add-process subcommand"""

# pylint: disable=missing-docstring
import datetime as dt
import platform
from argparse import ArgumentParser
from unittest import mock

from unittest_fixtures import requires

from gbp_ps.cli import add_process

from . import TestCase, factories, make_build_process, parse_args


@requires("repo", "gbp", "console")
class AddProcessTests(TestCase):
    """Tests for gbp add-process"""

    maxDiff = None

    @mock.patch("gbp_ps.cli.add_process.now")
    def test(self, mock_now: mock.Mock) -> None:
        now = mock_now.return_value = dt.datetime(2023, 11, 20, 17, 57, tzinfo=dt.UTC)
        proc = make_build_process(
            add_to_repo=False, build_host=platform.node(), start_time=now
        )
        console = self.fixtures.console
        cmdline = f"gbp add-process {proc.machine} {proc.build_id} {proc.package} {proc.phase}"
        args = parse_args(cmdline)
        exit_status = add_process.handler(args, self.fixtures.gbp, console)

        self.assertEqual(exit_status, 0)
        self.assertEqual([*self.fixtures.repo.get_processes()], [proc])

    def test_parse_args(self) -> None:
        # Just ensure that parse_args is there and works
        parser = ArgumentParser()
        add_process.parse_args(parser)


@requires("tempdb", "repo_fixture")
class AddProcessAddLocalProcessesTests(TestCase):
    def test(self) -> None:
        process = factories.BuildProcessFactory()

        add_process.add_local_process(self.fixtures.tempdb)(process)

        result = self.fixtures.repo.get_processes()

        self.assertEqual(list(result), [process])


class BuildProcessFromArgsTests(TestCase):
    def test(self) -> None:
        expected = factories.BuildProcessFactory()
        cmdline = (
            f"gbp add-process {expected.machine} {expected.build_id} {expected.package}"
            f" {expected.phase}"
        )
        args = parse_args(cmdline)

        with mock.patch("gbp_ps.cli.add_process.now", return_value=expected.start_time):
            with mock.patch(
                "gbp_ps.cli.add_process.platform.node", return_value=expected.build_host
            ):
                process = add_process.build_process_from_args(args)

        self.assertEqual(process, expected)



"""Tests for the exceptions module"""

# pylint: disable=missing-docstring
from typing import Any
from unittest import TestCase, mock

from gbp_ps.exceptions import RETURN_EXCEPTION, swallow_exception

from . import parametrized


class SwallowExceptionTests(TestCase):
    @parametrized(
        [(None, "test", None), (6, 6, Exception), (RETURN_EXCEPTION, "test", Exception)]
    )
    def test_swallow_exception(
        self, returns: Any, expected: Any, side_effect: type[BaseException]
    ) -> None:

        with mock.patch.object(self, "func", wraps=self.func, side_effect=side_effect):
            wrapped = swallow_exception(Exception, returns=returns)(self.func)
            result = wrapped()

        if returns is RETURN_EXCEPTION:
            self.assertIsInstance(result, side_effect)
        else:
            self.assertEqual(result, expected)

    @staticmethod
    def func() -> str:
        return "test"



"""add/update a process in the process table"""

import argparse
import datetime as dt
import platform
from functools import partial
from typing import Any, Callable, TypeAlias

from gbpcli.gbp import GBP
from gbpcli.graphql import check
from gbpcli.types import Console

from gbp_ps.repository import Repo, add_or_update_process
from gbp_ps.settings import Settings
from gbp_ps.types import BuildProcess

ProcessAdder: TypeAlias = Callable[[BuildProcess], Any]
now = partial(dt.datetime.now, tz=dt.UTC)


def handler(args: argparse.Namespace, gbp: GBP, _console: Console) -> int:
    """Show add/update an entry in the process table"""
    local: str | None = getattr(args, "local", None)
    add_process = add_local_process(local) if local else add_gbp_process(gbp)
    add_process(build_process_from_args(args))

    return 0


def add_gbp_process(gbp: GBP) -> ProcessAdder:
    """Return a function that can use GBP to add/update a given BuildProcess"""
    query = gbp.query.gbp_ps.add_process  # type: ignore[attr-defined]

    def add_process(process: BuildProcess) -> None:
        check(query(process=process.to_dict()))

    return add_process


def add_local_process(database: str) -> ProcessAdder:
    """Return a function that can use SqliteRepository to add/update a given BuildProcess"""
    repo = Repo(Settings(STORAGE_BACKEND="sqlite", SQLITE_DATABASE=database))

    def add_process(process: BuildProcess) -> None:
        add_or_update_process(repo, process)

    return add_process


def build_process_from_args(args: argparse.Namespace) -> BuildProcess:
    """Build and return a BuildProcess given the command-line args

    start_time will be the current time.
    build_host will be the current host.
    """
    return BuildProcess(
        build_host=platform.node(),
        build_id=args.number,
        machine=args.machine,
        package=args.package,
        phase=args.phase,
        start_time=now(),
    )


def parse_args(parser: argparse.ArgumentParser) -> None:
    """Set subcommand arguments"""
    parser.add_argument(
        "--local", "-l", default=None, help="(Where to) Use a local process database"
    )
    parser.add_argument("machine", metavar="MACHINE", help="name of the machine")
    parser.add_argument("number", metavar="NUMBER", help="build number")
    parser.add_argument("package", metavar="PACKAGE", help="package CPV")
    parser.add_argument("phase", metavar="PHASE", help="ebuild phase")



"""Show currently building packages"""

import argparse
import datetime as dt
import time
from typing import Any, Callable, NoReturn, TypeAlias

from gbpcli import render
from gbpcli.gbp import GBP
from gbpcli.graphql import check
from gbpcli.types import Console
from rich import box
from rich.console import RenderableType
from rich.live import Live
from rich.progress import BarColumn, Progress, TextColumn
from rich.table import Table

from gbp_ps import utils
from gbp_ps.exceptions import swallow_exception
from gbp_ps.repository import Repo
from gbp_ps.settings import Settings
from gbp_ps.types import BuildProcess

ProcessList: TypeAlias = list[BuildProcess]
ProcessGetter: TypeAlias = Callable[[], ProcessList]
ModeHandler = Callable[[argparse.Namespace, ProcessGetter, Console], int]

BUILD_PHASE_COUNT = len(BuildProcess.build_phases)
PHASE_PADDING = max(len(i) for i in BuildProcess.build_phases)


def handler(args: argparse.Namespace, gbp: GBP, console: Console) -> int:
    """Show currently building packages"""
    mode: ModeHandler = MODES[args.continuous]
    local: str | None = getattr(args, "local", None)

    get_processes = get_local_processes(local) if local else get_gbp_processes(gbp)

    return mode(args, get_processes, console)


def parse_args(parser: argparse.ArgumentParser) -> None:
    """Set subcommand arguments"""
    parser.add_argument(
        "--node", action="store_true", default=False, help="display the build node"
    )
    parser.add_argument(
        "--local", "-l", default=None, help="(Where to) Use a local process database"
    )
    parser.add_argument(
        "--continuous",
        "-c",
        action="store_true",
        default=False,
        help="Run and continuously poll and update",
    )
    parser.add_argument(
        "--update-interval",
        "-i",
        type=float,
        default=1,
        help="In continuous mode, the interval, in seconds, between updates",
    )
    parser.add_argument(
        "--progress",
        "-p",
        action="store_true",
        default=False,
        help="Display progress bars for package phase",
    )


def single_handler(
    args: argparse.Namespace, get_processes: ProcessGetter, console: Console
) -> int:
    """Handler for the single-mode run of `gbp ps`"""
    if processes := get_processes():
        console.out.print(create_table(processes, args))

    return 0


def get_gbp_processes(gbp: GBP) -> ProcessGetter:
    """Retrieve and return the ProcessList"""

    def get_processes() -> ProcessList:
        results = check(gbp.query.gbp_ps.get_processes())  # type: ignore[attr-defined]

        return [graphql_to_process(result) for result in results["buildProcesses"]]

    return get_processes


def get_local_processes(database: str) -> ProcessGetter:
    """Return a list of processes given the database path"""
    repo = Repo(Settings(STORAGE_BACKEND="sqlite", SQLITE_DATABASE=database))

    def get_processes() -> ProcessList:
        return list(repo.get_processes())

    return get_processes


@swallow_exception(KeyboardInterrupt, returns=0)
def continuous_handler(
    args: argparse.Namespace, get_processes: ProcessGetter, console: Console
) -> NoReturn:
    """Handler for the continuous-mode run of `gbp ps`"""

    def update() -> Table:
        return create_table(get_processes(), args)

    rate = 1 / args.update_interval
    out = console.out
    ctx = Live(update(), console=out, screen=out.is_terminal, refresh_per_second=rate)
    with ctx as live:
        while True:
            time.sleep(args.update_interval)
            live.update(update())


def graphql_to_process(result: dict[str, Any]) -> BuildProcess:
    """Return GraphQL build process output as BuildProcess object"""
    return BuildProcess(
        machine=result["machine"],
        build_id=result["id"],
        build_host=result["buildHost"],
        package=result["package"],
        phase=result["phase"],
        start_time=dt.datetime.fromisoformat(result["startTime"]),
    )


def create_table(processes: ProcessList, args: argparse.Namespace) -> Table:
    """Return a rich Table given the list of processes"""
    table = Table(
        title="Ebuild Processes",
        box=box.ROUNDED,
        expand=True,
        title_style="header",
        style="box",
    )
    table.add_column("Machine", header_style="header")
    table.add_column("ID", header_style="header")
    table.add_column("Package", header_style="header")
    table.add_column("Start", header_style="header")
    table.add_column("Phase", header_style="header")

    if args.node:
        table.add_column("Node", header_style="header")

    for process in processes:
        table.add_row(*row(process, args))

    return table


def row(process: BuildProcess, args: argparse.Namespace) -> list[RenderableType]:
    """Return a process row (list) given the process and args"""
    return [
        render.format_machine(process.machine, args),
        render.format_build_number(int(process.build_id)),
        f"[package]{process.package}[/package]",
        utils.format_timestamp(process.start_time.astimezone(render.LOCAL_TIMEZONE)),
        phase_column(process.phase, args),
        *([f"[build_host]{process.build_host}[/build_host]"] if args.node else []),
    ]


def phase_column(phase: str, args: argparse.Namespace) -> str | Progress:
    """Return the ebuild phase rendered for the process table column

    This will be the text of the ebuild phase and a progress bar depending on the
    args.progress flag and whether the phase is an ebuild build phase.
    """
    text = f"[{phase}_phase]{phase:{PHASE_PADDING}}[/{phase}_phase]"

    if not args.progress:
        return text

    position = utils.find(phase, BuildProcess.build_phases) + 1
    return progress(text, (position, BUILD_PHASE_COUNT) if position > 0 else None)


def progress(text: str, steps: tuple[int, int] | None) -> Progress:
    """Return Progress object with given text and steps (completed, total)

    If steps is None, a pulsing Progress bar is used.
    """
    prog = Progress(TextColumn(text), BarColumn())

    if steps is None:
        task = prog.add_task(text, total=None)
        return prog

    completed, total = steps
    task = prog.add_task(text, total=total)
    prog.update(task, advance=completed)
    return prog


MODES = [single_handler, continuous_handler]



"""Tests for gbp_ps.utils"""

# pylint: disable=missing-docstring
import datetime as dt
from unittest import TestCase, mock

from gbp_ps import utils


class FormatTimestampTests(TestCase):

    def test_when_today(self) -> None:
        timestamp = dt.datetime(2024, 2, 7, 20, 10)
        today = timestamp.date()

        with mock.patch("gbp_ps.utils.get_today", return_value=today):
            date_str = utils.format_timestamp(timestamp)

        self.assertEqual(date_str, "[timestamp]20:10:00[/timestamp]")

    def test_when_not_today(self) -> None:
        timestamp = dt.datetime(2024, 2, 7, 20, 10)
        today = (timestamp + dt.timedelta(hours=24)).date()

        with mock.patch("gbp_ps.utils.get_today", return_value=today):
            date_str = utils.format_timestamp(timestamp)

        self.assertEqual(date_str, "[timestamp]Feb07[/timestamp]")



"""CLI unit tests for gbp-ps ps subcommand"""

# pylint: disable=missing-docstring
import datetime as dt
from argparse import ArgumentParser
from functools import partial
from unittest import mock

from unittest_fixtures import requires

from gbp_ps.cli import ps
from gbp_ps.types import BuildProcess

from . import LOCAL_TIMEZONE, TestCase, factories, make_build_process, parse_args


@requires("gbp", "console")
class PSTests(TestCase):
    """Tests for gbp ps"""

    maxDiff = None

    @mock.patch("gbpcli.render.LOCAL_TIMEZONE", new=LOCAL_TIMEZONE)
    @mock.patch("gbp_ps.cli.ps.utils.get_today", new=lambda: dt.date(2023, 11, 15))
    def test(self) -> None:
        t = dt.datetime
        for cpv, phase, start_time in [
            ["sys-apps/portage-3.0.51", "postinst", t(2023, 11, 14, 16, 20, 0)],
            ["sys-apps/shadow-4.14-r4", "package", t(2023, 11, 15, 16, 20, 1)],
            ["net-misc/wget-1.21.4", "compile", t(2023, 11, 15, 16, 20, 2)],
        ]:
            make_build_process(package=cpv, phase=phase, start_time=start_time)
        cmdline = "gbp ps"
        args = parse_args(cmdline)
        console = self.fixtures.console

        exit_status = ps.handler(args, self.fixtures.gbp, console)

        self.assertEqual(exit_status, 0)
        expected = """\
                                    Ebuild Processes                                    
╭─────────────┬────────┬──────────────────────────────────┬─────────────┬──────────────╮
│ Machine     │ ID     │ Package                          │ Start       │ Phase        │
├─────────────┼────────┼──────────────────────────────────┼─────────────┼──────────────┤
│ babette     │ 1031   │ sys-apps/portage-3.0.51          │ Nov14       │ postinst     │
│ babette     │ 1031   │ sys-apps/shadow-4.14-r4          │ 15:20:01    │ package      │
│ babette     │ 1031   │ net-misc/wget-1.21.4             │ 15:20:02    │ compile      │
╰─────────────┴────────┴──────────────────────────────────┴─────────────┴──────────────╯
"""
        self.assertEqual(console.out.file.getvalue(), expected)

    @mock.patch("gbpcli.render.LOCAL_TIMEZONE", new=LOCAL_TIMEZONE)
    @mock.patch("gbp_ps.cli.ps.utils.get_today", new=lambda: dt.date(2024, 8, 16))
    def test_with_progress(self) -> None:
        t = dt.datetime
        for cpv, phase, start_time in [
            ["pipeline", "world", t(2024, 8, 16, 16, 20, 1)],
            ["sys-apps/shadow-4.14-r4", "package", t(2024, 8, 16, 16, 20, 1)],
            ["net-misc/wget-1.21.4", "compile", t(2024, 8, 16, 16, 20, 2)],
        ]:
            make_build_process(package=cpv, phase=phase, start_time=start_time)
        cmdline = "gbp ps --progress"
        args = parse_args(cmdline)
        console = self.fixtures.console

        exit_status = ps.handler(args, self.fixtures.gbp, console)

        self.assertEqual(exit_status, 0)
        expected = """\
                                    Ebuild Processes                                    
╭─────────┬──────┬─────────────────────────┬──────────┬────────────────────────────────╮
│ Machine │ ID   │ Package                 │ Start    │ Phase                          │
├─────────┼──────┼─────────────────────────┼──────────┼────────────────────────────────┤
│ babette │ 1031 │ pipeline                │ 14:20:01 │ world     ━━━━━━━━━━━━━━━━━━━━ │
│ babette │ 1031 │ sys-apps/shadow-4.14-r4 │ 14:20:01 │ package   ━━━━━━━━━━━━━━━      │
│ babette │ 1031 │ net-misc/wget-1.21.4    │ 14:20:02 │ compile   ━━━━━━━━━━           │
╰─────────┴──────┴─────────────────────────┴──────────┴────────────────────────────────╯
"""
        self.assertEqual(console.out.file.getvalue(), expected)

    @mock.patch("gbpcli.render.LOCAL_TIMEZONE", new=LOCAL_TIMEZONE)
    @mock.patch("gbp_ps.cli.ps.utils.get_today", new=lambda: dt.date(2023, 11, 15))
    def test_with_node(self) -> None:
        t = dt.datetime
        for cpv, phase, start_time in [
            ["sys-apps/portage-3.0.51", "postinst", t(2023, 11, 15, 16, 20, 0)],
            ["sys-apps/shadow-4.14-r4", "package", t(2023, 11, 15, 16, 20, 1)],
            ["net-misc/wget-1.21.4", "compile", t(2023, 11, 15, 16, 20, 2)],
        ]:
            make_build_process(package=cpv, phase=phase, start_time=start_time)
        cmdline = "gbp ps --node"
        args = parse_args(cmdline)
        console = self.fixtures.console
        exit_status = ps.handler(args, self.fixtures.gbp, console)

        self.assertEqual(exit_status, 0)
        expected = """\
                                    Ebuild Processes                                    
╭───────────┬───────┬─────────────────────────────┬────────────┬─────────────┬─────────╮
│ Machine   │ ID    │ Package                     │ Start      │ Phase       │ Node    │
├───────────┼───────┼─────────────────────────────┼────────────┼─────────────┼─────────┤
│ babette   │ 1031  │ sys-apps/portage-3.0.51     │ 15:20:00   │ postinst    │ jenkins │
│ babette   │ 1031  │ sys-apps/shadow-4.14-r4     │ 15:20:01   │ package     │ jenkins │
│ babette   │ 1031  │ net-misc/wget-1.21.4        │ 15:20:02   │ compile     │ jenkins │
╰───────────┴───────┴─────────────────────────────┴────────────┴─────────────┴─────────╯
"""
        self.assertEqual(console.out.file.getvalue(), expected)

    def test_from_install_to_pull(self) -> None:
        t = dt.datetime
        machine = "babette"
        build_id = "1031"
        package = "sys-apps/portage-3.0.51"
        build_host = "jenkins"
        orig_start = t(2023, 11, 15, 16, 20, 0)
        cmdline = "gbp ps --node"
        args = parse_args(cmdline)
        update = partial(
            make_build_process,
            machine=machine,
            build_id=build_id,
            package=package,
            build_host=build_host,
            start_time=orig_start,
            update_repo=True,
        )
        update(phase="world")

        # First compile it
        console = self.fixtures.console
        ps.handler(args, self.fixtures.gbp, console)

        self.assertEqual(
            console.out.file.getvalue(),
            """\
                                    Ebuild Processes                                    
╭───────────┬────────┬──────────────────────────────┬─────────┬─────────────┬──────────╮
│ Machine   │ ID     │ Package                      │ Start   │ Phase       │ Node     │
├───────────┼────────┼──────────────────────────────┼─────────┼─────────────┼──────────┤
│ babette   │ 1031   │ sys-apps/portage-3.0.51      │ Nov15   │ world       │ jenkins  │
╰───────────┴────────┴──────────────────────────────┴─────────┴─────────────┴──────────╯
""",
        )

        # Now it's done compiling
        update(phase="clean", start_time=orig_start + dt.timedelta(seconds=60))
        console.out.file.seek(0)
        console.out.file.truncate()
        ps.handler(args, self.fixtures.gbp, console)

        self.assertEqual(console.out.file.getvalue(), "")

        # Now it's being pulled by GBP on another node
        update(
            build_host="gbp",
            phase="pull",
            start_time=orig_start + dt.timedelta(seconds=120),
        )
        console.out.file.seek(0)
        console.out.file.truncate()
        ps.handler(args, self.fixtures.gbp, console)

        self.assertEqual(
            console.out.file.getvalue(),
            """\
                                    Ebuild Processes                                    
╭────────────┬────────┬────────────────────────────────┬─────────┬─────────────┬───────╮
│ Machine    │ ID     │ Package                        │ Start   │ Phase       │ Node  │
├────────────┼────────┼────────────────────────────────┼─────────┼─────────────┼───────┤
│ babette    │ 1031   │ sys-apps/portage-3.0.51        │ Nov15   │ pull        │ gbp   │
╰────────────┴────────┴────────────────────────────────┴─────────┴─────────────┴───────╯
""",
        )

    def test_empty(self) -> None:
        cmdline = "gbp ps"
        args = parse_args(cmdline)
        console = self.fixtures.console
        exit_status = ps.handler(args, self.fixtures.gbp, console)

        self.assertEqual(exit_status, 0)
        self.assertEqual(console.out.file.getvalue(), "")

    @mock.patch("gbpcli.render.LOCAL_TIMEZONE", new=LOCAL_TIMEZONE)
    @mock.patch("gbp_ps.cli.ps.time.sleep")
    @mock.patch("gbp_ps.cli.ps.utils.get_today", new=lambda: dt.date(2023, 11, 11))
    def test_continuous_mode(self, mock_sleep: mock.Mock) -> None:
        processes = [
            make_build_process(package=cpv, phase=phase)
            for cpv, phase in [
                ["sys-apps/portage-3.0.51", "postinst"],
                ["sys-apps/shadow-4.14-r4", "package"],
                ["net-misc/wget-1.21.4", "compile"],
            ]
        ]
        cmdline = "gbp ps -c -i4"
        args = parse_args(cmdline)
        console = self.fixtures.console

        gbp = mock.Mock()
        mock_graphql_resp = [process.to_dict() for process in processes]
        gbp.query.gbp_ps.get_processes.side_effect = (
            ({"buildProcesses": mock_graphql_resp}, None),
            KeyboardInterrupt,
        )
        exit_status = ps.handler(args, gbp, console)

        self.assertEqual(exit_status, 0)
        expected = """\
                                    Ebuild Processes                                    
╭─────────────┬────────┬──────────────────────────────────┬─────────────┬──────────────╮
│ Machine     │ ID     │ Package                          │ Start       │ Phase        │
├─────────────┼────────┼──────────────────────────────────┼─────────────┼──────────────┤
│ babette     │ 1031   │ sys-apps/portage-3.0.51          │ 05:20:52    │ postinst     │
│ babette     │ 1031   │ sys-apps/shadow-4.14-r4          │ 05:20:52    │ package      │
│ babette     │ 1031   │ net-misc/wget-1.21.4             │ 05:20:52    │ compile      │
╰─────────────┴────────┴──────────────────────────────────┴─────────────┴──────────────╯"""
        self.assertEqual(console.out.file.getvalue(), expected)
        mock_sleep.assert_called_with(4)


class PSParseArgsTests(TestCase):
    def test(self) -> None:
        # Just ensure that parse_args is there and works
        parser = ArgumentParser()
        ps.parse_args(parser)


@requires("tempdb", "repo_fixture")
class PSGetLocalProcessesTests(TestCase):
    def test_with_0_processes(self) -> None:
        p = ps.get_local_processes(self.fixtures.tempdb)()

        self.assertEqual(p, [])

    def test_with_1_process(self) -> None:
        process = factories.BuildProcessFactory()
        self.fixtures.repo.add_process(process)

        p = ps.get_local_processes(self.fixtures.tempdb)()

        self.assertEqual(p, [process])

    def test_with_multiple_processes(self) -> None:
        for _ in range(5):
            process = factories.BuildProcessFactory()
            self.fixtures.repo.add_process(process)

        self.assertEqual(len(ps.get_local_processes(self.fixtures.tempdb)()), 5)

    def test_with_final_processes(self) -> None:
        for phase in BuildProcess.final_phases:
            process = factories.BuildProcessFactory(phase=phase)
            self.fixtures.repo.add_process(process)

        self.assertEqual(len(ps.get_local_processes(self.fixtures.tempdb)()), 0)



# Generated by Django 4.2.7 on 2023-11-12 13:06

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="BuildProcess",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("machine", models.CharField(db_index=True, max_length=255)),
                ("build_id", models.CharField(max_length=255)),
                ("build_host", models.CharField(max_length=255)),
                ("package", models.CharField(max_length=255)),
                ("phase", models.CharField(max_length=255)),
                ("start_time", models.DateTimeField()),
            ],
            options={
                "unique_together": {("machine", "build_id", "build_host", "package")}
            },
        )
    ]



"""Helper utilities"""

import datetime as dt
from typing import Any, Sequence

from gbpcli.render import LOCAL_TIMEZONE


def get_today() -> dt.date:
    """Return today's date"""
    return dt.datetime.now().astimezone(LOCAL_TIMEZONE).date()


def format_timestamp(timestamp: dt.datetime) -> str:
    """Format the timestamp as a string

    Like render.from_timestamp(), but if the date is today's date then only display the
    time. If the date is not today's date then only return the date.
    """
    if (date := timestamp.date()) == get_today():
        return f"[timestamp]{timestamp.strftime('%X')}[/timestamp]"
    return f"[timestamp]{date.strftime('%b%d')}[/timestamp]"


def find(item: Any, items: Sequence[Any]) -> int:
    """Return the index of the first item in items

    If item is not found in items, return -1.
    """
    try:
        return items.index(item)
    except ValueError:
        return -1



"""Sqlite RepositoryType"""

import datetime as dt
import sqlite3
from contextlib import contextmanager
from typing import Generator, Iterable

from gbp_ps.exceptions import RecordAlreadyExists, RecordNotFoundError
from gbp_ps.settings import Settings
from gbp_ps.types import BuildProcess


class SqliteRepository:
    """Sqlite Based Repository"""

    final_phases_t = tuple(BuildProcess.final_phases)
    filter_phases = f"WHERE phase NOT IN ({','.join(['?' for _ in final_phases_t])})"
    row_names = "machine, build_id, build_host, package, phase, start_time"

    def __init__(self, settings: Settings) -> None:
        database: bytes | str = settings.SQLITE_DATABASE
        self._database = database
        self.init_db()

    def add_process(self, process: BuildProcess) -> None:
        """Add the given BuildProcess to the repository

        If the process already exists in the repo, RecordAlreadyExists is raised
        """
        # If this package exists in another build, remove it. This (usually) means the
        # other build failed
        query = """
            DELETE FROM ebuild_process
            WHERE build_id != ? AND machine = ? AND package = ?
        """
        with self.cursor() as cursor:
            cursor.execute(query, (process.build_id, process.machine, process.package))

        query = f"""
            INSERT INTO ebuild_process ({self.row_names})
            VALUES (?,?,?,?,?,?)
        """
        with self.cursor() as cursor:
            try:
                cursor.execute(query, self.process_to_row(process))
            except sqlite3.IntegrityError:
                raise RecordAlreadyExists(process) from None

    def update_process(self, process: BuildProcess) -> None:
        """Update the given build process

        Only updates the phase field

        If the build process doesn't exist in the repo, RecordNotFoundError is raised.
        """
        query = """
            SELECT machine,build_id,build_host,package,phase,start_time
            FROM ebuild_process
            WHERE machine = ? AND build_id = ? AND package = ?
        """
        with self.cursor() as cursor:
            result = cursor.execute(
                query, (process.machine, process.build_id, process.package)
            )
            row = result.fetchone()
        if not row:
            raise RecordNotFoundError(process) from None
        previous = self.row_to_process(*row)
        previous.ensure_updateable(process)
        query = """
            UPDATE ebuild_process
            SET phase = ?
            WHERE machine = ? AND build_id = ? AND package = ?
        """
        with self.cursor() as cursor:
            result = cursor.execute(
                query,
                (process.phase, process.machine, process.build_id, process.package),
            )

    def get_processes(self, include_final: bool = False) -> Iterable[BuildProcess]:
        """Return the process records from the repository

        If include_final is True also include processes in their "final" phase. The
        default value is False.
        """
        exclude_final = not include_final
        query = f"""
            SELECT machine,build_id,build_host,package,phase,start_time
            FROM ebuild_process
            {self.filter_phases if exclude_final else ""}
            ORDER BY start_time
        """
        params = self.final_phases_t if exclude_final else ()

        with self.cursor() as cursor:
            result = cursor.execute(query, params)
            for row in result:
                yield self.row_to_process(*row)

    @staticmethod
    def row_to_process(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        machine: str,
        build_id: str,
        build_host: str,
        package: str,
        phase: str,
        start_time: int,
    ) -> BuildProcess:
        """Return a BuildProcess given the sql row data"""
        return BuildProcess(
            machine=machine,
            build_id=build_id,
            build_host=build_host,
            package=package,
            phase=phase,
            start_time=dt.datetime.fromtimestamp(start_time, tz=dt.UTC),
        )

    @staticmethod
    def process_to_row(process: BuildProcess) -> tuple[str, str, str, str, str, int]:
        """Return the tuple of rows given the BuildProcess"""
        p = process
        start_time = int(p.start_time.timestamp())
        return (p.machine, p.build_id, p.build_host, p.package, p.phase, start_time)

    def init_db(self) -> None:
        """Initialize the database"""
        create_table = """
CREATE TABLE IF NOT EXISTS ebuild_process (
    machine VARCHAR(255),
    build_id VARCHAR(255),
    build_host VARCHAR(255),
    package VARCHAR(255),
    phase VARCHAR(255),
    start_time INTEGER
);
"""
        create_machine_idx = """
CREATE INDEX IF NOT EXISTS idx_machine
ON ebuild_process (machine)
"""
        create_phase_idx = """
CREATE INDEX IF NOT EXISTS idx_phase
ON ebuild_process (phase)
"""
        create_unique_idx = """
CREATE UNIQUE INDEX IF NOT EXISTS idx_unique_process
ON ebuild_process (machine, build_id, build_host, package)
"""
        with self.cursor() as cursor:
            cursor.execute(create_table)
            cursor.execute(create_machine_idx)
            cursor.execute(create_phase_idx)
            cursor.execute(create_unique_idx)

    @contextmanager
    def cursor(self) -> Generator[sqlite3.Cursor, None, None]:
        """Connect to the db and return a cursor object"""
        with sqlite3.connect(self._database) as connection:
            cursor = connection.cursor()
            try:
                yield cursor
            finally:
                cursor.close()



# Generated by Django 4.2.7 on 2023-11-12 18:51

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("gbp_ps", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="buildprocess",
            name="phase",
            field=models.CharField(db_index=True, max_length=255),
        )
    ]



"""GBP signal handlers for gbp-ps"""

import datetime as dt
import platform
from functools import cache, partial
from typing import Any

from gentoo_build_publisher.signals import dispatcher
from gentoo_build_publisher.types import Build

from gbp_ps.repository import Repo, RepositoryType, add_or_update_process
from gbp_ps.settings import Settings
from gbp_ps.types import BuildProcess

_now = partial(dt.datetime.now, tz=dt.UTC)
_NODE = platform.node()


def build_process(
    build: Build, build_host: str, phase: str, start_time: dt.datetime
) -> BuildProcess:
    """Return a BuildProcess with the given phase and timestamp"""
    return BuildProcess(
        build_host=build_host,
        build_id=build.build_id,
        machine=build.machine,
        package="pipeline",
        phase=phase,
        start_time=start_time,
    )


@cache
def repo() -> RepositoryType:
    """Return the Repository from from the environment variable settings"""
    return Repo(Settings.from_environ())


def set_process(build: Build, phase: str) -> None:
    """Add or update the given Build process in the repo"""
    add_or_update_process(repo(), build_process(build, _NODE, phase, _now()))


def prepull_handler(*, build: Build) -> None:
    """Signal handler for pre-pulls"""
    set_process(build, "pull")


def postpull_handler(*, build: Build, **_kwargs: Any) -> None:
    """Signal handler for post-pulls"""
    set_process(build, "clean")


dispatcher.bind(prepull=prepull_handler)
dispatcher.bind(postpull=postpull_handler)



"""gbp-ps data types"""

from __future__ import annotations

import datetime as dt
from dataclasses import asdict, dataclass
from typing import Any

from .exceptions import UpdateNotAllowedError


@dataclass(frozen=True, slots=True, kw_only=True)
class BuildProcess:
    """The basic build process type"""

    machine: str
    build_id: str
    build_host: str
    package: str
    phase: str
    start_time: dt.datetime

    # BuildProcesses in any of these phases are considered "final"
    final_phases = {"", "clean", "cleanrm", "postrm"}

    # Build phases, in order (best as I can determine)
    build_phases = (
        "pretend",
        "setup",
        "unpack",
        "prepare",
        "configure",
        "compile",
        "test",
        "install",
        "package",
        "instprep",
        "preinst",
        "postinst",
    )

    def is_same_as(self, other: BuildProcess) -> bool:
        """Return true if the other build process is the same process

        Two process are considered the "same" if the machine, package and build_id are
        the same.
        """
        return (
            self.package == other.package
            and self.machine == other.machine
            and self.build_id == other.build_id
        )

    def ensure_updateable(self, new: BuildProcess) -> None:
        """Raise an exception if process should not be updated to new"""
        if self.build_host != new.build_host and new.phase in BuildProcess.final_phases:
            raise UpdateNotAllowedError(self, new)

    def to_dict(self) -> dict[str, Any]:
        """Return BuildProcess as a GraphQL dict"""
        bp_dict = asdict(self)
        bp_dict["buildHost"] = bp_dict.pop("build_host")
        bp_dict["id"] = bp_dict.pop("build_id")
        bp_dict["startTime"] = bp_dict.pop("start_time").isoformat()

        return bp_dict

    def is_finished(self) -> bool:
        """Return True iff the BuildProcess is in a "final" phase"""
        return self.phase in self.final_phases



"""Tests for gbp-ps repositories"""

# pylint: disable=missing-docstring, duplicate-code
from dataclasses import replace
from typing import Any, Callable
from unittest import mock

import fakeredis
from unittest_fixtures import requires

from gbp_ps.exceptions import (
    RecordAlreadyExists,
    RecordNotFoundError,
    UpdateNotAllowedError,
)
from gbp_ps.repository import Repo, RepositoryType, add_or_update_process, sqlite
from gbp_ps.settings import Settings
from gbp_ps.types import BuildProcess

from . import TestCase, factories, make_build_process, parametrized

HOST = 0
REDIS_FROM_URL = "gbp_ps.repository.redis.redis.Redis.from_url"


def get_repo(backend: str, settings: Settings) -> RepositoryType:
    global HOST  # pylint: disable=global-statement
    settings = replace(settings, STORAGE_BACKEND=backend)
    if backend == "redis":
        fake_redis = fakeredis.FakeRedis(host=f"host{(HOST := HOST + 1)}")
        repo_patch = mock.patch(REDIS_FROM_URL, return_value=fake_redis)
    else:
        repo_patch = mock.MagicMock()

    with repo_patch:
        return Repo(settings)


def repos(*names: str) -> Callable[[Callable[[Any, str], None]], None]:
    return parametrized([[name] for name in names])


@requires("settings")
class RepositoryTests(TestCase):
    options = {
        "environ": {"GBP_PS_REDIS_KEY": "gbp-ps-test", "GBP_PS_KEY_EXPIRATION": "3600"}
    }

    @repos("django", "redis", "sqlite")
    def test_add_process(self, backend: str) -> None:
        repo = get_repo(backend, self.fixtures.settings)
        build_process: BuildProcess = factories.BuildProcessFactory()
        repo.add_process(build_process)
        self.assertEqual([*repo.get_processes()], [build_process])

    @repos("django", "redis", "sqlite")
    def test_add_process_when_already_exists(self, backend: str) -> None:
        repo = get_repo(backend, self.fixtures.settings)
        build_process: BuildProcess = factories.BuildProcessFactory()
        repo.add_process(build_process)

        with self.assertRaises(RecordAlreadyExists):
            repo.add_process(build_process)

    @repos("django", "redis", "sqlite")
    def test_add_process_same_package_in_different_builds_exist_only_once(
        self, backend: str
    ) -> None:
        repo = get_repo(backend, self.fixtures.settings)
        dead_process: BuildProcess = factories.BuildProcessFactory()
        repo.add_process(dead_process)
        new_process: BuildProcess = factories.BuildProcessFactory(
            machine=dead_process.machine,
            build_host=dead_process.build_host,
            package=dead_process.package,
        )
        repo.add_process(new_process)

        self.assertEqual([*repo.get_processes()], [new_process])

    @repos("django", "redis", "sqlite")
    def test_update_process(self, backend: str) -> None:
        repo = get_repo(backend, self.fixtures.settings)
        orig_process: BuildProcess = factories.BuildProcessFactory(phase="compile")
        repo.add_process(orig_process)

        updated_process = replace(orig_process, phase="postinst")

        repo.update_process(updated_process)

        expected = replace(orig_process, phase="postinst")
        self.assertEqual([*repo.get_processes()], [expected])

    @repos("django", "redis", "sqlite")
    def test_update_process_finalize_when_not_owned(self, backend: str) -> None:
        # This demonstrates the concept of build host "ownership". A a process can only
        # be updated with a "final" phase if the build host is the same. Otherwise it
        # should raise an exception
        repo = get_repo(backend, self.fixtures.settings)
        process1 = make_build_process(add_to_repo=False)
        repo.add_process(process1)
        process2 = replace(process1, build_host="badhost", phase="clean")

        with self.assertRaises(UpdateNotAllowedError):
            repo.update_process(process2)

    @repos("django", "redis")
    def test_add_or_update_process_can_handle_buildhost_changes(
        self, backend: str
    ) -> None:
        repo = get_repo(backend, self.fixtures.settings)
        orig_process: BuildProcess = factories.BuildProcessFactory(phase="clean")
        repo.add_process(orig_process)

        updated_process = replace(orig_process, build_host="gbp", phase="pull")

        add_or_update_process(repo, updated_process)

        expected = replace(orig_process, build_host="gbp", phase="pull")
        self.assertEqual([*repo.get_processes()], [expected])

    @repos("django", "redis", "sqlite")
    def test_add_or_update_ignores_notallowederror(self, backend: str) -> None:
        repo = get_repo(backend, self.fixtures.settings)
        process1 = make_build_process(add_to_repo=False)
        repo.add_process(process1)
        process2 = replace(process1, build_host="badhost", phase="clean")

        add_or_update_process(repo, process2)

        self.assertEqual([*repo.get_processes()], [process1])

    @repos("django", "redis", "sqlite")
    def test_update_process_when_process_not_in_db(self, backend: str) -> None:
        repo = get_repo(backend, self.fixtures.settings)
        build_process: BuildProcess = factories.BuildProcessFactory()

        with self.assertRaises(RecordNotFoundError):
            repo.update_process(build_process)

    @repos("django", "redis", "sqlite")
    def test_get_processes_with_empty_list(self, backend: str) -> None:
        repo = get_repo(backend, self.fixtures.settings)
        self.assertEqual([*repo.get_processes()], [])

    @repos("django", "redis", "sqlite")
    def test_get_processes_with_process(self, backend: str) -> None:
        repo = get_repo(backend, self.fixtures.settings)
        build_process: BuildProcess = factories.BuildProcessFactory()
        repo.add_process(build_process)

        self.assertEqual([*repo.get_processes()], [build_process])

    @repos("django", "redis", "sqlite")
    def test_get_processes_with_final_process(self, backend: str) -> None:
        repo = get_repo(backend, self.fixtures.settings)
        build_process: BuildProcess = factories.BuildProcessFactory(phase="postrm")
        repo.add_process(build_process)

        self.assertEqual([*repo.get_processes()], [])

    @repos("django", "redis", "sqlite")
    def test_get_processes_with_include_final_process(self, backend: str) -> None:
        repo = get_repo(backend, self.fixtures.settings)
        build_process: BuildProcess = factories.BuildProcessFactory(phase="postrm")
        repo.add_process(build_process)

        self.assertEqual([*repo.get_processes(include_final=True)], [build_process])

    def test_repo_factory_success(self) -> None:
        settings = replace(self.fixtures.settings, STORAGE_BACKEND="sqlite")
        repo = Repo(settings)

        self.assertTrue(isinstance(repo, sqlite.SqliteRepository))

    def test_repo_factory_failure(self) -> None:
        settings = replace(self.fixtures.settings, STORAGE_BACKEND="bogus")

        with self.assertRaises(ValueError):
            Repo(settings)



"""Settings for tests"""

from pathlib import Path

BASE_DIR = Path(__file__).resolve().parent.parent

SECRET_KEY = "test"

USE_TZ = True

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "gentoo_build_publisher.apps.GentooBuildPublisherConfig",
    "gbp_ps.apps.GBPPSConfig",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

DATABASES = {
    "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": BASE_DIR / "db.sqlite3"}
}

ROOT_URLCONF = "gentoo_build_publisher.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]

CELERY_TASK_ALWAYS_EAGER = True



"""gbp-ps settings"""

from dataclasses import dataclass
from typing import ClassVar

from gbpcli.settings import BaseSettings

DEFAULT_REDIS_KEY_EXPIRATION = 3600 * 24


@dataclass(frozen=True, slots=True)
class Settings(BaseSettings):
    """Settings for gbp-ps"""

    # pylint: disable=invalid-name
    env_prefix: ClassVar = "GBP_PS_"

    REDIS_KEY: str = "gbp-ps"
    REDIS_KEY_EXPIRATION: int = DEFAULT_REDIS_KEY_EXPIRATION
    REDIS_URL: str = "redis://redis.invalid:6379/0"
    SQLITE_DATABASE: str = ":memory:"
    STORAGE_BACKEND: str = "django"



"""GraphQL interface for gbp-ps"""

# pylint: disable=missing-docstring
from importlib import resources
from typing import Any

from ariadne import ObjectType, convert_kwargs_to_snake_case, gql
from graphql import GraphQLResolveInfo

from gbp_ps.repository import Repo, add_or_update_process
from gbp_ps.settings import Settings
from gbp_ps.types import BuildProcess

type_defs = gql(resources.read_text("gbp_ps", "schema.graphql"))
resolvers = [
    build_process := ObjectType("BuildProcess"),
    query := ObjectType("Query"),
    mutation := ObjectType("Mutation"),
]


ADD_BUILD_FIELDS = {"machine", "id", "package", "phase"}


@build_process.field("id")
def resolve_build_process_id(process: BuildProcess, _info: GraphQLResolveInfo) -> str:
    return process.build_id


@query.field("buildProcesses")
@convert_kwargs_to_snake_case
def resolve_query_build_processes(
    _obj: Any, _info: GraphQLResolveInfo, *, include_final: bool = False
) -> list[BuildProcess]:
    """Return the list of BuildProcesses

    If include_final is True also include processes in their "final" phase. The default
    value is False.
    """
    return list(
        Repo(Settings.from_environ()).get_processes(include_final=include_final)
    )


@mutation.field("addBuildProcess")
@convert_kwargs_to_snake_case
def resolve_mutation_add_build_process(
    _obj: Any, _info: GraphQLResolveInfo, process: dict[str, Any]
) -> None:
    """Add the given process to the process table

    If the process already exists in the table, it is updated with the new value
    """
    # Don't bother when required fields are empty.
    if not all(process[field] for field in ADD_BUILD_FIELDS):
        return

    process["build_id"] = process.pop("id")
    add_or_update_process(Repo(Settings.from_environ()), BuildProcess(**process))



"""CLI unit tests for gbp-ps add-process subcommand"""

# pylint: disable=missing-docstring
import argparse
from unittest import mock

from unittest_fixtures import requires

from gbp_ps.cli import dump_bashrc

from . import TestCase


@requires("gbp", "console")
class DumpBashrcHandlerTests(TestCase):
    def test_without_local(self) -> None:
        args = argparse.Namespace(url="http://gbp.invalid/", local=False)
        fixtures = self.fixtures
        console = fixtures.console

        exit_status = dump_bashrc.handler(args, fixtures.gbp, console)

        self.assertEqual(exit_status, 0)

        lines = console.out.file.getvalue().split("\n")
        self.assertTrue(lines[0].startswith("if [[ -f /Makefile.gbp"))
        self.assertTrue("http://gbp.invalid/graphql" in lines[-4], lines[-4])

    def test_local(self) -> None:
        args = argparse.Namespace(url="http://gbp.invalid/", local=True)
        fixtures = self.fixtures
        console = fixtures.console
        tmpdir = "/var/bogus"

        with mock.patch("gbp_ps.cli.dump_bashrc.sp.Popen") as popen:
            process = popen.return_value.__enter__.return_value
            process.wait.return_value = 0
            process.stdout.read.return_value = tmpdir.encode("utf-8")
            exit_status = dump_bashrc.handler(args, fixtures.gbp, console)

        self.assertEqual(exit_status, 0)

        output = console.out.file.getvalue()
        self.assertTrue(f"{tmpdir}/portage/gbpps.db" in output, output)

    def test_local_portageq_fail(self) -> None:
        args = argparse.Namespace(url="http://gbp.invalid/", local=True)
        fixtures = self.fixtures
        console = fixtures.console

        with mock.patch("gbp_ps.cli.dump_bashrc.sp.Popen") as popen:
            process = popen.return_value.__enter__.return_value
            process.wait.return_value = 1
            exit_status = dump_bashrc.handler(args, fixtures.gbp, console)

        self.assertEqual(exit_status, 0)

        output = console.out.file.getvalue()
        self.assertTrue("/var/tmp/portage/gbpps.db" in output, output)


class ParseArgsTests(TestCase):
    def test(self) -> None:
        # Just ensure that parse_args is there and works
        parser = argparse.ArgumentParser()
        dump_bashrc.parse_args(parser)



#!/usr/bin/env python
"""Run tests for Gentoo Build Publisher"""
import os
import sys

import django
from django.conf import settings
from django.test.utils import get_runner


def main() -> None:
    """Program entry point"""
    os.environ["DJANGO_SETTINGS_MODULE"] = "tests.settings"

    # These values are required in order to import the publisher module
    os.environ.setdefault("BUILD_PUBLISHER_JENKINS_BASE_URL", "http://jenkins.invalid/")
    os.environ.setdefault("BUILD_PUBLISHER_STORAGE_PATH", "__testing__")

    django.setup()

    tests = sys.argv[1:] or ["."]

    TestRunner = get_runner(settings)  # pylint: disable=invalid-name
    test_runner = TestRunner(failfast=True, verbosity=2)
    failures = test_runner.run_tests(tests)

    sys.exit(bool(failures))


if __name__ == "__main__":
    main()



"""AppConfigs for gbp-ps"""

from django.apps import AppConfig


class GBPPSConfig(AppConfig):
    """AppConfig for gbp-ps"""

    name = "gbp_ps"
    verbose_name = "GBP-ps"
    default_auto_field = "django.db.models.BigAutoField"

    def ready(self) -> None:
        """Django app initialization"""
        # pylint: disable=import-outside-toplevel,unused-import
        # register signal handlers
        import gbp_ps.signals
